
/*
** Copyright (c) 2006 Herbert Poetzl
**
** V0.01	initial prototype
** V0.02	multi cpu support
**
*/

#define VERSION "V0.02"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>

#include "vserver.h"
#include "switch.h"

#include "history.h"
#include "monitor.h"
#include "debug_cmd.h"

static	char	opt_id = 0;
static	char	opt_vers = 0;
static	char	opt_cpus = 0;
static	char	opt_hist = 0;
static	char	opt_disass = 0;

static	char *	cmd_name = NULL;

static	int	num_id = 0;
static	int	num_ret = 0;
static	int	num_cpus = 1;
static	int	num_buf = 1024;
static	int	num_delay = 1000;

static uint32_t *cpuidx;


int	read_monitor(int cpu, struct vcmd_read_monitor_v0 *vc_data)
{
	return vserver(VCMD_read_monitor, cpu, vc_data);
}

int	read_history(int cpu, struct vcmd_read_history_v0 *vc_data)
{
	return vserver(VCMD_read_history, cpu, vc_data);
}

struct	_dump_hdr {
	uint32_t index;
	uint16_t count;
	uint16_t cpu;
};


void	dump_monitor(int cpu, struct vcmd_read_monitor_v0 *vc_data)
{
	struct _dump_hdr hdr = { vc_data->index, vc_data->count, cpu };

	fprintf(stderr, "[%u] %5u entries @%08x\n", cpu,
	    vc_data->count, vc_data->index);
	write(1, &hdr, sizeof(hdr));
	write(1, vc_data->data, sizeof(*vc_data->data) * vc_data->count);
}

void	dump_history(int cpu, struct vcmd_read_history_v0 *vc_data)
{
	struct _dump_hdr hdr = { vc_data->index, vc_data->count, cpu };

	fprintf(stderr, "[%u] %5u entries @%08x\n", cpu,
	    vc_data->count, vc_data->index);
	write(1, &hdr, sizeof(hdr));
	write(1, vc_data->data, sizeof(*vc_data->data) * vc_data->count);
}


#define OPTIONS "hdi:B:D:HI:V"

int	main(int argc, char *argv[])
{
	extern int optind;
	extern char *optarg;
	char c, errflg = 0;
	int r, cpu;

	cmd_name = argv[0];
	while ((c = getopt(argc, argv, "+" OPTIONS)) != EOF) {
	    switch (c) {
	    case 'h':
		fprintf(stderr,
		    "This is %s " VERSION "\n"
		    "options are:\n"
		    "  -h        print this help message\n"
		    "  -d        disassemble data\n"
		    "  -i <num>  cpu id/start\n"
		    "  -B <num>  buffer size\n"
		    "  -D <num>  poll delay (ms)\n"
		    "  -H        dump history\n"
		    "  -I <num>  cpus to dump\n"
		    "  -V        verify interface version\n"
		    "  --        end of options\n"
		    ,cmd_name);
		exit(0);
		break;
	    case 'd':
		opt_disass = 1;
		break;
	    case 'i':
		num_id = atoi(optarg);
		opt_id = 1;
		break;
	    case 'B':
		num_buf = atoi(optarg);
		break;
	    case 'D':
		num_delay = atoi(optarg);
		break;
	    case 'H':
		opt_hist = 1;
		break;
	    case 'I':
		num_cpus = atoi(optarg);
		opt_cpus = 1;
		break;
	    case 'V':
		opt_vers = 1;
		break;
	    case '?':
	    default:
		errflg++;
		break;
	    }
	}
	if (errflg) {
	    fprintf(stderr, 
		"Usage: %s [-" OPTIONS "] <args> [-e cmd ...]\n"
		"%s -h for help.\n",
		cmd_name, cmd_name);
	    exit(2);
	}

	if (opt_vers) {
	    r = vserver(VCMD_get_version, 0, NULL);
	    if (r<0)
		perror("vc_get_version");
	    else 
		printf("version: %04x:%04x\n", 
		    (r>>16) & 0xFFFF, r & 0xFFFF);
	    exit(0);
	}

	cpuidx = calloc(num_cpus, sizeof(*cpuidx));
	if (!cpuidx) {
	    perror("calloc");
	    exit(2);
	}

	if (opt_hist) {
	    struct _vx_hist_entry *data;
	    struct vcmd_read_history_v0 vc_data = { 0,0, NULL };

	    data = calloc(num_buf, sizeof(*data));
	    if (!data) {
		perror("calloc");
		exit(2);
	    }
	    vc_data.data = (char *)data;
	    fprintf(stderr, "[*] sync ...\n");
	    for (cpu = num_id; cpu < (num_id + num_cpus); cpu++) {
		vc_data.count = 0;
	        r = read_history(cpu, &vc_data);
		if (r<0)
		    continue;
		cpuidx[cpu - num_id] = vc_data.index;
	        dump_history(cpu, &vc_data);
	    }

	    while (1) {
		int sleep = 1;

		for (cpu = num_id; cpu < (num_id + num_cpus); cpu++) {
		    vc_data.count = num_buf;
		    vc_data.index = cpuidx[cpu - num_id];
		    r = read_history(cpu, &vc_data);
		    if (!vc_data.count)
			continue;
		    cpuidx[cpu - num_id] = vc_data.index;
		    dump_history(cpu, &vc_data);
		    
		    if (r<0) {
		        perror("read_history");
		        num_ret = 3;
		        goto out_hist;	
		    }
		    if (!!r)
			sleep = 0;
		}
		if (sleep) {
		    fprintf(stderr, "[*] sleeping ...\n");
		    usleep(num_delay * 1000);
		}
	    }
	out_hist:
	    free(data);

	} else {
	    struct _vx_mon_entry *data;
	    struct vcmd_read_monitor_v0 vc_data = { 0,0, NULL };

	    data = calloc(num_buf, sizeof(*data));
	    if (!data) {
		perror("calloc");
		exit(2);
	    }
	    vc_data.data = (char *)data;
	    fprintf(stderr, "[*] sync ...\n");
	    for (cpu = num_id; cpu < (num_id + num_cpus); cpu++) {
		vc_data.count = 0;
	        r = read_monitor(cpu, &vc_data);
		if (r<0)
		    continue;
		cpuidx[cpu - num_id] = vc_data.index;
	        dump_monitor(cpu, &vc_data);
	    }

	    while (1) {
		int sleep = 1;

		for (cpu = num_id; cpu < (num_id + num_cpus); cpu++) {
		    vc_data.count = num_buf;
		    vc_data.index = cpuidx[cpu - num_id];
		    r = read_monitor(cpu, &vc_data);
		    if (!vc_data.count)
			continue;
		    cpuidx[cpu - num_id] = vc_data.index;
		    dump_monitor(cpu, &vc_data);

		    if (r<0) {
			perror("read_monitor");
			num_ret = 3;
			goto out_mon;	
		    }
		}
		if (sleep) {
		    fprintf(stderr, "[*] sleeping ...\n");
		    usleep(num_delay * 1000);
		}
	    }
	out_mon:
	    free(data);
	}
	free(cpuidx);

	exit(num_ret);
}

